import datetime
import inspect
import json
import os
import pprint
import sys
import subprocess
import typing
import uuid

import pytray.pretty
from pytray import tree
import bson


def obj_dict(obj):
    """Given an object return a dictionary that represents it"""
    repr_dict = {}
    for name in dir(obj):
        if not name.startswith('_'):
            try:
                value = getattr(obj, name)
                if not inspect.isroutine(value):
                    repr_dict[name] = value
            except Exception as exc:  # pylint: disable=broad-except
                repr_dict[name] = '{}: {}'.format(type(exc).__name__, exc)

    return repr_dict


class UUIDEncoder(json.JSONEncoder):

    def default(self, obj):  # pylint: disable=arguments-differ
        if isinstance(obj, uuid.UUID):
            # if the obj is uuid, we simply return the value of uuid
            return repr(obj)
        if isinstance(obj, bson.ObjectId):
            return repr(obj)
        return json.JSONEncoder.default(self, obj)


class UUIDDecoder(json.JSONDecoder):

    def decode(self, s):  # pylint: disable=arguments-differ
        decoded = super(UUIDDecoder, self).decode(s)

        def to_uuid(entry, path):  # pylint: disable=unused-argument
            if isinstance(entry, str):
                if entry.startswith('UUID('):
                    try:
                        return uuid.UUID(entry[6:-2])
                    except ValueError:
                        pass
                elif entry.startswith('ObjectId('):
                    try:
                        return bson.ObjectId(entry[6:-2])
                    except ValueError:
                        pass

            return entry

        return tree.transform(to_uuid, decoded)


def pretty_type_string(obj_type: typing.Type) -> str:
    """Given an type will return a simple type string"""
    type_str = pytray.pretty.type_string(obj_type)
    parts = type_str.split('.')
    if len(parts) > 2:
        return "|".join([parts[0], parts[-1]])

    return type_str


def pretty_format(value, single_line=False, max_length=None) -> str:
    if isinstance(value, type):
        return pretty_type_string(value)
    if isinstance(value, datetime.datetime):
        if value.year == datetime.datetime.now().year:
            fmt = "%b %d %H:%M:%S"
        else:
            fmt = "%b %d %Y %H:%M:%S"

        return value.strftime(fmt)

    if isinstance(value, str):
        string = value
    else:
        string = pprint.pformat(value, depth=2, indent=0, compact=True)
    if single_line:
        string = string.replace('\n', '')
    if max_length and len(string) > max_length:
        string = string[:max_length - 3] + "..."

    return string


def open_file(filename):
    """Open a generic file on in a semi-portable way"""
    if sys.platform == "win32":
        os.startfile(filename)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, filename])



import logging
from typing import MutableSequence, Sequence

import stevedore

from . import plugins

__all__ = ('ActionManager',)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

ACTIONERS_NAMESPACE = 'mincepy_gui.actioners'


def get_actioners() -> Sequence:
    """Get all mincepy types and type helper instances registered as extensions"""
    mgr = stevedore.extension.ExtensionManager(
        namespace=ACTIONERS_NAMESPACE,
        invoke_on_load=False,
    )

    all_types = []

    def get_actioner(extension: stevedore.extension.Extension):
        try:
            all_types.extend(extension.plugin())
        except Exception:  # pylint: disable=broad-except
            logger.exception("Failed to get actioner plugin from %s", extension.name)

    mgr.map(get_actioner)

    return all_types


class ActionManager:
    """Group all actioners together and load plugins"""

    def __init__(self):
        self._actioners = []  # type: MutableSequence[plugins.Actioner]

    def load_plugins(self):
        self._actioners.extend(get_actioners())

    def register(self, actioner: plugins.Actioner):
        self._actioners.append(actioner)

    def probe(self, obj, context) -> Sequence:
        actions = []

        for actioner in self._actioners:
            possible_actions = actioner.probe(obj, context)
            if possible_actions:
                for action in possible_actions:
                    actions.append((action, actioner))

        return actions

    def get_actioners(self, type=None, name: str = None) -> Sequence[plugins.Actioner]:  # pylint: disable=redefined-builtin
        actioners = []
        for actioner in self._actioners:
            # Filter by all the criteria
            if type is not None and not isinstance(actioner, type):
                continue
            if name is not None and getattr(actioner, 'name', None) != name:
                continue

            actioners.append(actioner)

        return actioners



from typing import Iterable

import mincepy_gui
import mincepy_gui.actioners


class TestActioner(mincepy_gui.Actioner):

    def __init__(self, actions_list: list):
        self._actions_list = actions_list
        self._trigger_count = {}

    def probe(self, obj, context) -> Iterable[str]:
        return self._actions_list

    def do(self, action, obj, context):
        self._trigger_count[action] = self._trigger_count.get(action, 0) + 1


if __name__ == "__main__":
    """A manual way to test.  Not ideal but better than nothing"""
    test_actioner = mincepy_gui.actioners.TestActioner
    test_actioner.enabled = True
    test_actioner.actions = ('test-action1', 'test-action2')
    mincepy_gui.start()



from collections import namedtuple
from functools import partial
import logging
import typing

import PySide2
from PySide2 import QtCore, QtGui
from PySide2.QtCore import QObject, Signal, Slot, Qt, QModelIndex
from pytray import tree

import mincepy
from . import common
from . import utils

__all__ = 'DbModel', 'SnapshotRecord'

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

UNSET = ''  # The value printed for records that don't have a particular attribute

TOOLTIPS = {
    mincepy.TYPE_ID: 'Object type',
    mincepy.CREATION_TIME: 'Creation time',
    mincepy.SNAPSHOT_TIME: 'Last modification time',
    mincepy.VERSION: 'Version',
}

SnapshotRecord = namedtuple("SnapshotRecord", 'snapshot record')


class DbModel(QObject):
    # Signals
    historian_changed = Signal(mincepy.Historian)

    def __init__(self):
        super().__init__()
        self._historian = None

    @property
    def historian(self) -> mincepy.Historian:
        return self._historian

    @Slot(mincepy.Historian)
    def set_historian(self, historian):
        self._historian = historian
        self.historian_changed.emit(self._historian)


class DataRecordQueryModel(QtCore.QAbstractTableModel):
    # Signals
    type_restriction_changed = Signal(object)
    sort_changed = Signal(dict)
    query_changed = Signal(dict)

    NE_DELETED = {'$ne': mincepy.DELETED}

    def __init__(self, db_model: DbModel, executor=common.default_executor, parent=None):
        super().__init__(parent)
        self._db_model = db_model
        self._query = {'state': self.NE_DELETED}
        self._results = None
        self._sort = None
        self._type_restriction = None
        self._column_names = mincepy.DataRecord._fields
        self._update_future = None

        self._executor = executor
        self._new_results.connect(self._inject_results)

        # If the historian changes then we get invalidated
        self._db_model.historian_changed.connect(lambda hist: self._invalidate_results())

    @property
    def db_model(self):
        return self._db_model

    @property
    def column_names(self):
        return self._column_names

    def get_query(self):
        return self._query

    def set_query(self, query: dict):
        """Set the query that will be passed to historian.find()"""
        if query == self._query:
            return

        self._query = query
        self.query_changed.emit(query)
        self._invalidate_results()

    def update_query(self, update: dict):
        new_query = self.get_query().copy()
        new_query.update(update)
        self.set_query(new_query)

    def set_type_restriction(self, type_id):
        if type_id == self.get_type_restriction():
            return

        restriction = {'obj_type': type_id}
        self.update_query(restriction)
        self.type_restriction_changed.emit(self.get_type_restriction())

    def get_type_restriction(self):
        return self.get_query().get('obj_type', None)

    def set_show_deleted(self, value):
        if self.get_show_deleted() == value:
            return

        no_deleted = {'state': mincepy.DELETED}

        query = self.get_query()
        if value:
            if 'state' in query:
                state = query['state']
                if isinstance(state, list):
                    if no_deleted in state:
                        state.remove(no_deleted)
                        self.set_query(query)
                else:
                    if state == no_deleted:
                        query.pop('state')
                        self.set_query(query)
        else:
            pass

    def get_show_deleted(self) -> bool:
        no_deleted = {'$ne': mincepy.DELETED}
        query = self.get_query()
        if 'state' in query:
            state = query['state']
            if isinstance(state, list):
                return no_deleted in state

            return no_deleted == state

        return False

    def set_sort(self, sort):
        if sort == self.get_sort():
            return

        self.update_query({'sort': sort})
        self.sort_changed.emit(self.get_sort())

    def get_sort(self):
        return self._query.get('sort', None)

    def get_records(self):
        return self._results

    def refresh(self):
        self._invalidate_results()

    def rowCount(self, _parent: QtCore.QModelIndex = QModelIndex()) -> int:
        if self._results is None:
            return 0

        return len(self._results)

    def columnCount(self, _parent: QtCore.QModelIndex = QModelIndex()) -> int:
        return len(self.column_names)

    def headerData(self,
                   section: int,
                   orientation: PySide2.QtCore.Qt.Orientation,
                   role: int = ...) -> typing.Any:
        if role != Qt.DisplayRole:
            return None

        if orientation == QtCore.Qt.Orientation.Horizontal:
            return self.column_names[section]

        return str(self._results[section])

    def data(self, index: PySide2.QtCore.QModelIndex, role: int = ...) -> typing.Any:
        if role == Qt.DisplayRole:
            value = getattr(self._results[index.row()], self.column_names[index.column()])
            return str(value)

        return None

    def _update_results(self):
        if self._query is None or self._db_model.historian is None:
            self._results = []
        else:
            self._results = []
            self._update_future = self._executor(partial(self._perform_query, self.get_query()),
                                                 msg="Querying...",
                                                 blocking=False)

    def _invalidate_results(self):
        self.beginResetModel()
        self._results = None
        self.endResetModel()
        self._update_results()

    def _perform_query(self, query, batch_size=4):
        logging.debug("Starting query: %s", query)

        total = 0
        batch = []
        for result in self._db_model.historian.find_records(**query):
            batch.append(result)
            if len(batch) == batch_size:
                self._new_results.emit(batch)
                batch = []
            total += 1

        if batch:
            # Emit the last batch
            self._new_results.emit(batch)

        logging.debug('Finished query, got %s results', total)

    _new_results = Signal(list)

    @QtCore.Slot(list)
    def _inject_results(self, batch: list):
        """As a query is executed batches of results as emitted and passed to this callback for
        insertion"""
        first = len(self._results)
        last = first + len(batch) - 1
        self.beginInsertRows(QModelIndex(), first, last)
        self._results.extend(batch)
        self.endInsertRows()


class EntriesTable(QtCore.QAbstractTableModel):
    DEFAULT_COLUMNS = (mincepy.TYPE_ID, mincepy.CREATION_TIME, mincepy.SNAPSHOT_TIME,
                       mincepy.VERSION, mincepy.STATE)

    object_activated = Signal(object)

    def __init__(self, query_model: DataRecordQueryModel, parent=None):
        super().__init__(parent)
        self._query_model = query_model
        self._query_model.modelReset.connect(self._invalidate)
        self._query_model.rowsInserted.connect(self._query_rows_inserted)

        self._columns = list(self.DEFAULT_COLUMNS)
        self._show_objects = True

        self._snapshots_cache = {}

    def get_records(self):
        return self._query_model.get_records()

    @property
    def query_model(self):
        return self._query_model

    def get_record(self, row) -> typing.Optional[mincepy.DataRecord]:
        if row < 0 or row >= self.rowCount():
            return None

        return self.get_records()[row]

    def get_snapshot(self, row: int):
        ref = self.get_record(row).get_reference()
        if ref not in self._snapshots_cache:
            historian = self._query_model.db_model.historian
            try:
                self._snapshots_cache[ref] = historian.load_snapshot(ref)
            except TypeError as exc:
                logger.info("Failed to load snapshot for reference '%s'\n%s", ref, exc)
                self._snapshots_cache[ref] = None

        return self._snapshots_cache[ref]

    def get_show_as_objects(self):
        return self._show_objects

    def set_show_as_objects(self, as_objects):
        if self._show_objects != as_objects:
            self._show_objects = as_objects
            # Remove the old columns, if there are any to be removed
            if len(self._columns) > len(self.DEFAULT_COLUMNS):
                first = len(self.DEFAULT_COLUMNS)
                last = len(self._columns) - 1

                self.beginRemoveColumns(QModelIndex(), first, last)
                self._reset_columns()
                self.endRemoveColumns()

            # Now add the new ones
            columns = set()
            for row in range(self.rowCount()):
                columns.update(self._get_columns_for(row))

            if columns:
                # Convert to list and sort alphabetically
                new_columns = list(columns)
                new_columns.sort()
                self._insert_columns(new_columns)

    def refresh(self):
        # Ask the query to refresh from the database
        self._query_model.refresh()

    def rowCount(self, _parent: PySide2.QtCore.QModelIndex = ...) -> int:
        return self._query_model.rowCount()

    def columnCount(self, _parent: PySide2.QtCore.QModelIndex = ...) -> int:
        return len(self._columns)

    def headerData(self,
                   section: int,
                   orientation: PySide2.QtCore.Qt.Orientation,
                   role: int = ...) -> typing.Any:
        if role != Qt.DisplayRole:
            return None

        if orientation == QtCore.Qt.Orientation.Horizontal:
            if section >= len(self._columns):
                return None
            return self._columns[section]

        if orientation == QtCore.Qt.Orientation.Vertical:
            return str(section)

        return None

    def data(self, index: PySide2.QtCore.QModelIndex, role: int = ...) -> typing.Any:
        column_name = self._columns[index.column()]
        if role == common.DataRole:
            return self._get_value(index.row(), index.column())
        if role == Qt.DisplayRole:
            return self._get_value_string(index.row(), index.column())
        if role == Qt.FontRole:
            if column_name in mincepy.DataRecord._fields:
                font = QtGui.QFont()
                font.setItalic(True)
                return font
        if role == Qt.ToolTipRole:
            try:
                return TOOLTIPS[column_name]
            except KeyError:
                pass

        return None

    def sort(self, column: int, order: PySide2.QtCore.Qt.SortOrder = ...):
        column_name = self._columns[column]
        try:
            sort_criterion = column_name
        except KeyError:
            if self._show_objects:
                # We can't deal with sorting objects at the moment
                return
            sort_criterion = "state.{}".format(column_name)

        sort_dict = {
            sort_criterion: mincepy.ASCENDING if order == Qt.AscendingOrder else mincepy.DESCENDING
        }
        self._query_model.set_sort(sort_dict)

    @Slot(QModelIndex)
    def activate_entry(self, index: QModelIndex):
        obj = self.data(index, role=common.DataRole)
        if obj is not None and obj != UNSET:
            self.object_activated.emit(obj)

    def _invalidate(self):
        self.beginResetModel()
        self._snapshots_cache = {}
        self._reset_columns()
        self.endResetModel()

    def _get_snapshot_state(self, record):
        if self._show_objects:
            historian = self._query_model.db_model.historian
            try:
                return historian.load_snapshot(record.get_reference())
            except TypeError:
                pass  # Fall back to displaying the state

        return record.state

    def _get_columns_for(self, row: int) -> tuple:
        if self.get_show_as_objects():
            obj = self.get_snapshot(row)
            try:
                return utils.obj_dict(obj).keys()
            except TypeError:
                pass
        else:
            state = self.get_record(row).state
            if isinstance(state, dict):
                return tuple("state.{}".format(key) for key in state)

        return ()

    def _get_value(self, row: int, column: int) -> typing.Any:
        """Get a value for a given row and column index"""
        column_name = self._columns[column]
        if not column_name:
            return UNSET

        column_path = column_name.split('.')
        record = self.get_record(row)

        if column_path[0] in record._fields:
            try:
                record_value = tree.get_by_path(record._asdict(), column_path)
            except (KeyError, IndexError, TypeError):
                # Probably trying to access the state using path from a different object type
                return UNSET

            # Special case to show type ids as the class name
            if column_name == mincepy.TYPE_ID:
                try:
                    historian = self._query_model.db_model.historian
                    return historian.get_obj_type(record_value)
                except TypeError:
                    pass
            return record_value

        # The column is a custom attribute of the item
        if self.get_show_as_objects():
            snapshot = self.get_snapshot(row)
            try:
                return utils.obj_dict(snapshot).get(column_name, UNSET)
            except TypeError:
                pass

        return UNSET

    def _get_value_string(self, row: int, column: int) -> str:
        value = self._get_value(row, column)
        return utils.pretty_format(value, single_line=True, max_length=100)

    def _query_rows_inserted(self, _parent: QModelIndex, first: int, last: int):
        """Called when there are new entries inserted into the entries table"""

        self.beginInsertRows(QModelIndex(), first, last)

        columns = set()  # Keep track of all the columns in this batch
        for row in range(first, last + 1):
            columns.update(self._get_columns_for(row))

        self.endInsertRows()

        # Check if the columns need updating
        columns -= set(self._columns)
        if columns:
            # There are new columns to insert
            cols = list(columns)
            cols.sort()
            self._insert_columns(cols)

    def _insert_columns(self, new_columns: typing.Sequence):
        """Add new columns to our existing ones"""
        first_col = len(self._columns)
        last_col = first_col + len(new_columns) - 1
        self.beginInsertColumns(QModelIndex(), first_col, last_col)
        self._columns.extend(new_columns)
        self.endInsertColumns()

    def _reset_columns(self):
        """Reset the columns back to the default internally.  This should only be done between
        either a model invalidation or appropriate removeColumns call
        """
        self._columns = list(self.DEFAULT_COLUMNS)



from concurrent.futures import Future

from PySide2.QtCore import Qt
from pytray.futures import capture_exceptions

__all__ = ('DataRole',)


def default_executor(func, msg=None, blocking=False):  # pylint: disable=unused-argument
    future = Future()
    with capture_exceptions(future):
        future.set_result(func())

    return future


# Role to get the actual data associated with an index
DataRole = Qt.UserRole  # pylint: disable=invalid-name



from pathlib import Path
from typing import Iterable, Optional
import tempfile

import mincepy

from . import plugins
from . import utils


class TextActioner(plugins.Actioner):

    @property
    def name(self):
        return "text-actioner"

    def probe(self, obj, context) -> Optional[Iterable[str]]:
        if isinstance(obj, mincepy.File):
            return ('View File',)

        return None

    def do(self, action: str, obj: mincepy.File, context: dict):
        if action == 'View File':
            folder = Path(tempfile.mkdtemp())
            obj.to_disk(folder)
            utils.open_file(folder / obj.filename)


class DataRecordActioner(plugins.Actioner):
    """Knows how to perform actions on data records"""

    @property
    def name(self):
        return "data-record-actioner"

    def probe(self, obj, context) -> Optional[Iterable[str]]:
        if isinstance(obj, mincepy.DataRecord):
            return ("Delete",)
        if isinstance(obj, Iterable) \
                and all(map(lambda val: isinstance(val, mincepy.DataRecord), obj)):
            count = len(tuple(obj))
            return ("Delete {}".format(count),)

        return None

    def do(self, action, obj, context):
        pass


class CopyActioner(plugins.Actioner):
    """Knows how to copy things to the clipboard"""

    @property
    def name(self):
        return "copy-actioner"

    def probe(self, obj, context: dict) -> Optional[Iterable[str]]:
        actions = []
        if isinstance(obj, mincepy.DataRecord):
            actions.append("Copy Object ID")
        elif hasattr(obj, '__str__'):
            actions.append("Copy")

        return actions

    def do(self, action, obj, context: dict):
        clipboard = context.get('clipboard', None)
        if isinstance(obj, mincepy.DataRecord):
            clipboard.setText(str(obj.obj_id))
        else:
            clipboard.setText(str(obj))


class TestActioner(plugins.Actioner):
    enabled = False
    actions = []

    @property
    def name(self):
        return "test-actioner"

    def probe(self, obj, context: dict) -> Optional[Iterable[str]]:
        if self.enabled:
            return self.actions

    def do(self, action, obj, context: dict):
        print("Action '{}' called with object: \n'{}' and context \n'{}'".format(
            action, obj, context))


def get_actioners():
    return (CopyActioner(), TextActioner(), DataRecordActioner(), TestActioner())



from functools import partial
import json

from PySide2.QtCore import QObject, Signal, Slot, QPoint
from PySide2 import QtWidgets, QtGui
import mincepy

from . import common
from . import models
from . import tree_models
from . import utils


class DatabaseController(QObject):
    """Controls the connection to the database"""

    _historian_created = Signal(mincepy.Historian)

    # pylint: disable=too-many-arguments
    def __init__(self,
                 db_model: models.DbModel,
                 uri_line: QtWidgets.QLineEdit,
                 connect_button: QtWidgets.QPushButton,
                 default_uri='',
                 executor=common.default_executor,
                 parent=None):
        super().__init__(parent)
        self._db_model = db_model
        self._uri_line = uri_line
        self._connect_button = connect_button
        self._executor = executor

        self._uri_line.setText(default_uri)

        self._historian_created.connect(self._db_model.set_historian)
        self._connect_button.clicked.connect(self._handle_connect)
        self._uri_line.returnPressed.connect(self._handle_connect)

    def _handle_connect(self):
        uri = self._uri_line.text()
        self._executor(partial(self._connect, uri), "Connecting", blocking=True)

    def _connect(self, uri):
        try:
            historian = mincepy.create_historian(uri)
        except Exception as exc:
            err_msg = "Error creating historian with uri '{}':\n{}".format(uri, exc)
            raise RuntimeError(err_msg)
        else:
            self._historian_created.emit(historian)

        return "Connected to {}".format(uri)


class QueryController(QObject):
    """Controls the query that is being applied on the database"""

    def __init__(self,
                 query_model: models.DataRecordQueryModel,
                 query_line: QtWidgets.QLineEdit,
                 parent=None):
        super().__init__(parent)
        self._query_model = query_model
        self._query_line = query_line

        def set_query_edited():
            palette = self._query_line.palette()
            palette.setColor(palette.Base, QtGui.QColor(192, 212, 192))
            query_line.setPalette(palette)

        def reset_query_edited():
            palette = self._query_line.palette()
            palette.setColor(palette.Base, QtGui.QColor(255, 255, 255))
            query_line.setPalette(palette)

        def handle_query_changed(new_query: dict):
            new_text = self._query_to_str(new_query)
            if new_text != self._query_line.text():
                self._query_line.setText(new_text)
            reset_query_edited()

        def handle_query_submitted():
            new_text = self._query_line.text()
            try:
                query = json.loads(new_text, cls=utils.UUIDDecoder)
            except json.decoder.JSONDecodeError as exc:
                QtWidgets.QErrorMessage().showMessage(str(exc))
            else:
                self._query_model.set_query(query)

        def handle_text_edited(_text):
            try:
                current_query = json.dumps(self._query_model.get_query(), cls=utils.UUIDEncoder)
            except json.decoder.JSONDecodeError:
                pass
            else:
                if _text != current_query:
                    set_query_edited()
                else:
                    reset_query_edited()

        self._query_model.query_changed.connect(handle_query_changed)
        self._query_line.returnPressed.connect(handle_query_submitted)
        self._query_line.textEdited.connect(handle_text_edited)

        self._query_line.setText(self._query_to_str(self._query_model.get_query()))

    @staticmethod
    def _query_to_str(query: dict) -> str:
        return json.dumps(query, cls=utils.UUIDEncoder)


class EntryDetailsController(QObject):
    """Controller that set what is displayed in the details tree"""

    context_menu_requested = Signal(dict, QPoint)

    def __init__(self,
                 entries_table: models.EntriesTable,
                 entries_table_view: QtWidgets.QTableView,
                 entry_details_view: QtWidgets.QTreeWidget,
                 entry_details: tree_models.RecordTree = None,
                 parent=None):
        super().__init__(parent)
        self._entries_table = entries_table
        self._entries_table_view = entries_table_view
        self._entry_details_view = entry_details_view
        self._entry_details = entry_details or tree_models.RecordTree(self)

        # Configure the view
        self._entry_details_view.setContextMenuPolicy(QtGui.Qt.CustomContextMenu)
        self._entry_details_view.customContextMenuRequested.connect(self._entry_context_menu)

        def handle_row_changed(current, _previous):
            record = self._entries_table.get_record(current.row())
            snapshot = self._entries_table.get_snapshot(current.row())
            self._entry_details.set_record(record, snapshot)

        self._entries_table_view.selectionModel().currentRowChanged.connect(handle_row_changed)

    def handle_copy(self, copier: callable):
        objects = self._get_currently_selected_objects()
        if not objects:
            return

        objects = objects if len(objects) > 1 else objects[0]
        copier(objects)

    @Slot(QPoint)
    def _entry_context_menu(self, point: QPoint):
        objects = self._get_currently_selected_objects()
        if objects:
            groups = {}
            groups['Objects'] = objects if len(objects) > 1 else objects[0]
            self.context_menu_requested.emit(groups, self._entry_details_view.mapToGlobal(point))

    def _get_currently_selected_objects(self):
        # We just want the value, we don't care about the other columns
        value_col = tree_models.RecordTree.COLUMN_HEADERS.index('Value')
        selected = tuple(
            index for index in self._entry_details_view.selectionModel().selectedIndexes()
            if index.column() == value_col)

        return tuple(self._entry_details.data(index, role=common.DataRole) for index in selected)


class EntriesTableController(QObject):
    """Controller for the table showing database entries"""
    DATA_RECORDS = 'Data Record(s)'
    VALUES = 'Values(s)'

    context_menu_requested = Signal(dict, QPoint)

    def __init__(self,
                 entries_table: models.EntriesTable,
                 entries_table_view: QtWidgets.QTableView,
                 parent=None):
        """
        :param entries_table: the entries table model
        :param entries_table_view: the entries table view
        :param parent: the parent widget
        """
        super().__init__(parent)
        self._entries_table = entries_table
        self._entries_table_view = entries_table_view

        # Configure the view
        self._entries_table_view.setContextMenuPolicy(QtGui.Qt.CustomContextMenu)
        self._entries_table_view.customContextMenuRequested.connect(self._entries_context_menu)

    def get_selected(self) -> dict:
        groups = {}
        selected = self._entries_table_view.selectionModel().selectedIndexes()

        rows = {index.row() for index in selected}
        rows = sorted(tuple(rows))
        data_records = tuple(self._entries_table.get_record(row) for row in rows)

        if data_records:
            groups[self.DATA_RECORDS] = data_records if len(data_records) > 1 else data_records[0]

        objects = tuple(self._entries_table.data(index, role=common.DataRole) for index in selected)
        if objects:
            groups[self.VALUES] = objects if len(objects) > 1 else objects[0]

        return groups

    def handle_copy(self, copier: callable):
        if not copier:
            return

        selected = self._entries_table_view.selectionModel().selectedIndexes()
        if selected:
            rows = {index.row() for index in selected}
            data_records = tuple(self._entries_table.get_record(row) for row in rows)
            # Convert to scalar if needed
            data_records = data_records if len(data_records) > 1 else data_records[0]
            copier(data_records)

    @Slot(QPoint)
    def _entries_context_menu(self, point: QPoint):
        groups = self.get_selected()
        self.context_menu_requested.emit(groups, self._entries_table_view.mapToGlobal(point))



from functools import partial
from typing import Mapping, Iterable

from PySide2.QtCore import QObject, QPoint, Slot
from PySide2 import QtWidgets

from . import common
from . import extend

__all__ = ('CONTEXT_CLIPBOARD',)

CONTEXT_CLIPBOARD = 'clipboard'


class ActionController(QObject):

    def __init__(self,
                 action_manager: extend.ActionManager,
                 context=None,
                 executor=common.default_executor,
                 parent=None):
        super().__init__(parent)
        self._action_manager = action_manager
        self._context = context or {}
        self._executor = executor

    @Slot(dict, QPoint)
    def trigger_context_menu(self, groups: Mapping[str, Iterable], where: QPoint):
        """Request a context menu at the given qpoint that can respond to the objects given in the
        passed groups.  Groups should be a dictionary that maps a group name (str) to an iterable
        of objects to be acted on by actioners."""
        menu = QtWidgets.QMenu()
        for group, obj in groups.items():
            group_created = False
            actions = self._action_manager.probe(obj, self._context)
            for name, actioner in actions:
                if not group_created:
                    menu.addSection(group)
                    group_created = True

                menu.addAction(name, partial(actioner.do, name, obj, self._context))
        if menu.actions():
            menu.exec_(where)



from .action_controllers import *
from .common import *
from .extend import *
from .version import *
from .main import *
from .plugins import *
from . import controllers

__all__ = action_controllers.__all__ + common.__all__ + version.__all__ + main.__all__ + \
          plugins.__all__ + extend.__all__



from concurrent.futures import ThreadPoolExecutor, Future
from functools import partial
import logging

from PySide2 import QtWidgets
from PySide2 import QtGui
from PySide2.QtCore import QObject, Qt, Slot, Signal

from . import action_controllers
from . import controllers
from . import extend
from . import models
from . import tree_models
from . import types_controller

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class MainController(QObject):

    def __init__(self, window, default_uri=''):
        super().__init__(window)
        self._window = window
        self._default_uri = default_uri
        self._executor = ThreadPoolExecutor()
        self._tasks = []
        self._action_manager = extend.ActionManager()
        self._action_context = {
            action_controllers.CONTEXT_CLIPBOARD: QtGui.QGuiApplication.clipboard()
        }
        self._copier = None
        self._load_plugins()

        # Models
        self._create_models()

        # Views
        self._init_views(window)

        # Controllers
        self._create_controllers(window)

        self._init_shortcuts()
        self._task_done_signal.connect(self._task_done)
        self._status_bar.showMessage('Ready')

    def _load_plugins(self):
        logger.info("Starting loading plugins")
        self._action_manager.load_plugins()
        actioners = self._action_manager.get_actioners(name='copy-actioner')
        if actioners:
            # Somewhat arbitrarily just use the last one
            self._copier = partial(actioners[-1].do, 'copy', context=self._action_context)

        logger.info("Finished loading plugins")

    def _init_shortcuts(self):
        ctrl_c = QtGui.QKeySequence("Ctrl+C")

        QtWidgets.QShortcut(ctrl_c, self._window.splitter, self._copy)

    @Slot()
    def _copy(self):
        if not self._copier:
            return

        if self._window.entries_table.hasFocus():
            self._entries_table_controller.handle_copy(self._copier)
        elif self._window.entry_details.hasFocus():
            self._entry_details_controller.handle_copy(self._copier)

    def _create_models(self):
        self._db_model = models.DbModel()
        self._query_model = models.DataRecordQueryModel(self._db_model, self._execute, parent=self)
        self._entries_table = models.EntriesTable(self._query_model, parent=self)
        self._entry_details = tree_models.RecordTree(parent=self)

    def _init_views(self, window):
        window.entries_table.setSortingEnabled(True)
        window.entries_table.setModel(self._entries_table)
        window.entry_details.setModel(self._entry_details)

        self._init_display_as_class(window)
        window.refresh_button.clicked.connect(self._entries_table.refresh)

        # Keep out status bar
        self._status_bar = window.status_bar  # type: QtWidgets.QStatusBar

    def _init_display_as_class(self, window):
        window.display_as_class.stateChanged.connect(
            lambda state: self._entries_table.set_show_as_objects(state == Qt.Checked))
        self._entries_table.set_show_as_objects(window.display_as_class.checkState() == Qt.Checked)

    def _create_controllers(self, window):
        controllers.DatabaseController(self._db_model,
                                       window.uri_line,
                                       window.connect_button,
                                       default_uri=self._default_uri,
                                       executor=self._execute,
                                       parent=self)
        controllers.QueryController(self._query_model, window.query_line, parent=self)

        action_controller = action_controllers.ActionController(self._action_manager,
                                                                context=self._action_context,
                                                                executor=self._executor,
                                                                parent=self)

        # Entries table
        self._entries_table_controller = controllers.EntriesTableController(self._entries_table,
                                                                            window.entries_table,
                                                                            parent=self)
        self._entries_table_controller.context_menu_requested.connect(
            action_controller.trigger_context_menu)

        # Entry details
        self._entry_details_controller = controllers.EntryDetailsController(self._entries_table,
                                                                            window.entries_table,
                                                                            window.entry_details,
                                                                            self._entry_details,
                                                                            parent=self)
        self._entry_details_controller.context_menu_requested.connect(
            action_controller.trigger_context_menu)

        types_controller.TypeFilterController(self._query_model, window.type_filter, parent=self)

    def _execute(self, func, msg=None, blocking=False) -> Future:
        future = self._executor.submit(func)
        QtWidgets.QApplication.setOverrideCursor(Qt.WaitCursor if blocking else Qt.BusyCursor)
        self._tasks.append(future)
        future.add_done_callback(self._task_done_signal.emit)
        if msg is not None:
            self._status_bar.showMessage(msg)

        return future

    _task_done_signal = Signal(Future)

    @Slot(object)
    def _object_activated(self, obj):
        if self._app_common.type_viewers:
            self._execute(partial(self._app_common.call_viewers, obj),
                          msg="Calling viewer",
                          blocking=True)

    @Slot(Future)
    def _task_done(self, future):
        self._tasks.remove(future)
        QtWidgets.QApplication.restoreOverrideCursor()

        if not self._tasks:
            self._status_bar.clearMessage()

        try:
            new_msg = future.result()
            if new_msg is not None:
                self._status_bar.showMessage(new_msg, 1000)
        except Exception as exc:  # pylint: disable=broad-except
            QtWidgets.QErrorMessage(self._window).showMessage(str(exc))



from abc import ABCMeta, abstractmethod
from typing import Iterable, Optional

__all__ = ('Actioner',)


class Actioner(metaclass=ABCMeta):

    @abstractmethod
    def probe(self, obj, context: dict) -> Optional[Iterable[str]]:
        """Return an iterable of action names this actioner can perform given the object and
        the context"""

    @abstractmethod
    def do(self, action, obj, context: dict):  # pylint: disable=invalid-name
        """Perform the action on the given object in the given context"""



# File: main.py

from pathlib import Path
import sys

from PySide2.QtUiTools import QUiLoader
from PySide2 import QtWidgets
from PySide2.QtCore import QFile, Qt
from PySide2 import QtGui

from . import main_controllers

__all__ = ('start',)

RESOURCES = Path(__file__).parent / 'res'


def start(default_uri=''):
    if hasattr(Qt, 'AA_ShareOpenGLContexts'):
        QtWidgets.QApplication.setAttribute(Qt.AA_ShareOpenGLContexts)

    app = QtWidgets.QApplication(sys.argv)

    ui_file = QFile(str(RESOURCES / "mainwindow.ui"))
    ui_file.open(QFile.ReadOnly)

    loader = QUiLoader()
    window = loader.load(ui_file)
    ui_file.close()

    # Add the icon
    window.setWindowIcon(QtGui.QIcon(str(RESOURCES / "logo.svg")))

    main_controllers.MainController(window, default_uri)
    window.show()

    sys.exit(app.exec_())


if __name__ == "__main__":
    start()



from PySide2.QtCore import QObject
from PySide2.QtWidgets import QComboBox, QCompleter
from PySide2 import QtWidgets

from . import common
from . import models
from . import utils


class TypeFilterController(QObject):
    """Drop down combo box that lists the types available in archive"""
    ALL = None

    def __init__(self,
                 query_model: models.DataRecordQueryModel,
                 view: QtWidgets.QComboBox = None,
                 executor=common.default_executor,
                 parent=None):
        super().__init__(parent)
        self._view = self._configure_view(view or QComboBox(self))
        self._query_model = query_model
        self._executor = executor

        query_model.db_model.historian_changed.connect(self._update)

        self._types = [None]

    def _configure_view(self, view):
        view.setEditable(True)
        view.addItem(self.ALL)

        def handle_index_changed(index):
            if index >= 0:
                restrict_type = self._types[index]
                self._query_model.set_type_restriction(restrict_type)

        view.currentIndexChanged.connect(handle_index_changed)

        return view

    @property
    def _historian(self):
        return self._query_model.db_model.historian

    def _update(self):
        self._view.clear()
        self._executor(self._gather_types, "Gathering types", blocking=False)

    def _gather_types(self):
        results = self._historian.archive.find()
        self._types = [None]
        self._types.extend(list(set(result.type_id for result in results)))

        type_names = self._get_type_names(self._types)

        self._view.addItems(type_names)
        completer = QCompleter(type_names)
        self._view.setCompleter(completer)

    def _get_type_names(self, types):
        type_names = []
        for type_id in types:
            try:
                helper = self._historian.get_helper(type_id)
            except TypeError:
                type_names.append(str(type_id))
            else:
                type_names.append(utils.pretty_type_string(helper.TYPE))

        return type_names



# -*- coding: utf-8 -*-
from setuptools import setup

__author__ = "Martin Uhrin"
__license__ = "GPLv3 and MIT"
__contributors__ = "Sonia Collaud"

about = {}
with open('mincepy_gui/version.py') as f:
    exec(f.read(), about)

setup(
    name='mincepy-gui',
    version=about['__version__'],
    description="GUI for navigating your mincepy database visually!",
    long_description=open('README.rst').read(),
    url='https://github.com/muhrin/mincepy_gui.git',
    author='Martin Uhrin',
    author_email='martin.uhrin.10@ucl.ac.uk',
    license=__license__,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='database schemaless nosql object-store gui',
    install_requires=[
        'mincepy>=0.12.0',
        'PySide2',
        'pytray>=0.2.2, <0.3.0',
        'stevedore',
    ],
    extras_require={
        'gui': [],
        'dev': [
            'pip',
            'pytest>4',
            'pytest-cov',
            'pytest-qt',
            'pre-commit',
            'yapf',
            'prospector',
            'pylint',
            'twine',
        ],
    },
    packages=['mincepy_gui'],
    include_package_data=True,
    test_suite='test',
    provides=[],
    entry_points={'mincepy_gui.actioners': ['native_types = mincepy_gui.actioners:get_actioners',]})



from mincepy.testing import *
from mincepy_gui import models, tree_models

import mincepy_gui


def test_record_tree(historian):
    car = Car()
    car.save()
    record = historian.get_current_record(car)

    record_tree = tree_models.RecordTree()
    record_tree.set_record(record, car)
    # Should have the car and the record
    data = []
    data.append(
        record_tree.data(record_tree.index(0, tree_models.RecordTree.COLUMN_HEADERS.index('Value')),
                         role=mincepy_gui.DataRole))
    data.append(
        record_tree.data(record_tree.index(1, tree_models.RecordTree.COLUMN_HEADERS.index('Value')),
                         role=mincepy_gui.DataRole))
    assert car in data



"""Module for tree model related methods and classes"""
from abc import ABCMeta, abstractmethod
import operator
import typing
from typing import Sequence, Mapping

import PySide2
from PySide2 import QtCore
from PySide2.QtCore import QModelIndex, Qt, Signal, Slot

import mincepy
from . import common
from . import utils


class BaseTreeItem(metaclass=ABCMeta):

    def __init__(self, column_data: Sequence, parent=None):
        self._data = column_data
        self._strings = tuple(
            utils.pretty_format(datum, single_line=True, max_length=300) for datum in column_data)
        self._parent = parent

    def data(self, column: int):
        """Get the data at the given column"""
        if column < 0 or column >= self.column_count():
            return None

        return self._data[column]

    def string(self, column: int):
        """Get the string representation for the given column"""
        if column < 0 or column >= self.column_count():
            return None

        return self._strings[column]

    def row(self) -> int:
        """Get the row of this item within its parent"""
        if self._parent is None:
            return 0

        return self._parent.child_row(self)

    def column_count(self) -> int:
        """Get the number of columns in this item"""
        return len(self._data)

    def parent(self):
        """Get the parent of this tree item"""
        return self._parent

    @abstractmethod
    def child(self, row: int):
        """Get the child at the given row"""

    @abstractmethod
    def child_count(self) -> int:
        """Get the number of children this item has"""

    @abstractmethod
    def child_row(self, tree_item) -> int:
        """Get the row number given a child item"""


class DataTreeItem(BaseTreeItem):
    """Tree item that directly sores the required data internally"""

    def __init__(self, column_data: Sequence, parent=None):
        super().__init__(column_data, parent)
        self._children = []

    def append_child(self, child):
        self._children.append(child)

    def child(self, row: int):
        if row < 0 or row >= self.child_count():
            return None

        return self._children[row]

    def child_count(self) -> int:
        return len(self._children)

    def child_row(self, tree_item) -> int:
        return self._children.index(tree_item)


class LazyMappingItem(BaseTreeItem):

    def __init__(self, column_data: Sequence, raw_data, child_builder, num_children, parent=None):
        super(LazyMappingItem, self).__init__(column_data, parent)
        self._raw_data = raw_data
        self._child_builder = child_builder
        self._num_children = num_children
        self._children = {}

    def child_count(self) -> int:
        return self._num_children

    def child(self, row: int):
        if row < 0 or row >= self.child_count():
            return None

        if row not in self._children:
            self._children[row] = self._child_builder(self._raw_data, row, self)
            if len(self._children) == self._num_children:
                # Can discard the raw data and builder now
                self._raw_data = None
                self._child_builder = None

        return self._children[row]

    def child_row(self, tree_item) -> int:
        for row, child in self._children.items():
            if tree_item is child:
                return row
        raise ValueError("'{}' is not a child".format(tree_item))


class RecordTree(QtCore.QAbstractItemModel):
    COLUMN_HEADERS = 'Property', 'Type', 'Value'

    object_activated = Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._root_item = DataTreeItem(self.COLUMN_HEADERS)
        self._data_record = None  # type: typing.Optional[mincepy.DataRecord]

    @Slot(QModelIndex)
    def activate_entry(self, index: QModelIndex):
        obj = self.data(index, role=common.DataRole)
        if obj:
            self.object_activated.emit(obj)

    def index(self,
              row: int,
              column: int,
              parent: PySide2.QtCore.QModelIndex = QModelIndex()) -> \
            PySide2.QtCore.QModelIndex:
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parent_item = self._root_item
        else:
            parent_item = parent.internalPointer()

        child_item = parent_item.child(row)
        if child_item is not None:
            return self.createIndex(row, column, child_item)

        return QModelIndex()

    def parent(self, child: QModelIndex) -> QModelIndex:
        if not child.isValid():
            return QModelIndex()

        child_item = child.internalPointer()  # type: DataTreeItem
        parent_item = child_item.parent()  # type: DataTreeItem

        if parent_item is self._root_item:
            return QModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def rowCount(self, parent: PySide2.QtCore.QModelIndex = ...) -> int:
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parent_item = self._root_item
        else:
            parent_item = parent.internalPointer()

        return parent_item.child_count()

    def columnCount(self, _parent: PySide2.QtCore.QModelIndex = ...) -> int:
        return self._root_item.column_count()

    def data(self, index: PySide2.QtCore.QModelIndex, role: int = ...) -> typing.Any:
        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            item = index.internalPointer()  # type: BaseTreeItem
            return item.string(index.column())
        if role == common.DataRole:
            return index.internalPointer().data(index.column())

        return None

    def flags(self, index: PySide2.QtCore.QModelIndex) -> PySide2.QtCore.Qt.ItemFlags:
        if not index.isValid():
            return Qt.NoItemFlags

        return super(RecordTree, self).flags(index)

    def headerData(self,
                   section: int,
                   orientation: PySide2.QtCore.Qt.Orientation,
                   role: int = ...) -> typing.Any:
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._root_item.data(section)

        return None

    def set_record(self, record: mincepy.DataRecord, obj=None):
        """Set the data to visualise, the object instance can optionally be provided"""
        self.beginResetModel()
        self._data_record = record
        # Now build the three
        if self._data_record is None:
            self._root_item = DataTreeItem(self.COLUMN_HEADERS)
        else:
            tree_dict = {'record': record._asdict()}
            if obj is not None:
                tree_dict['obj'] = obj

            self._root_item = LazyMappingItem(self.COLUMN_HEADERS, tree_dict, self._item_builder,
                                              len(tree_dict))
        self.endResetModel()

    def _item_builder(self, build_from, row, parent=None):
        if isinstance(build_from, Sequence):
            key = str(row)
            try:
                child = build_from[row]
            except Exception as exc:  # pylint: disable=broad-except
                child = "Error getting child: {}".format(exc)
        elif isinstance(build_from, Mapping):
            entry = sorted(build_from.items(), key=operator.itemgetter(0))[row]
            key, child = str(entry[0]), entry[1]
        else:
            raise TypeError("Type '{}' does not support children")

        column_data = (key, utils.pretty_type_string(type(child)), child)

        # Is the child nested?
        nested_child_data = None
        if isinstance(child, str):
            pass
        elif isinstance(child, (Sequence, Mapping)):
            # We have a sequence or mapping
            nested_child_data = child
        else:
            try:
                nested_child_data = utils.obj_dict(child)
            except TypeError:
                pass

        if nested_child_data is not None:
            # We have a nested child so get a lazy item
            return LazyMappingItem(column_data, nested_child_data, self._item_builder,
                                   len(nested_child_data), parent)

        # Fall back to a plain unnested item
        return DataTreeItem(column_data, parent)
